#include <KernelExport.h>
#include <sys/ioctl.h>
#include <time.h>
#include <string.h>
#include <lzo1x.h>

#include "fsproto.h"
#include "netbfs.h"
#include "dircache.h"


#define DRIVER_NAME "netbfs"
#define ID "\033[34m" DRIVER_NAME ":\033[m "

#define DEBUG_DRIVER 1

#if DEBUG_DRIVER
#define DPRINTF(d,x) if (d) dprintf x
#else
#define DPRINTF(d,x) ((void)0)
#endif


#define NETBFS_ROOTNODE_ID 1

typedef struct replyinfo_t {
	uint8 status;
	uint32 data_length;
	uint8 continuation;
	char *data;
} replyinfo_t;

typedef struct iobuff_t {
	char *decomp_temp;
	char *recv_buff;
	char *send_buff;
} iobuff_t;


// node information
typedef struct fs_node {
	vnode_id id;
	vnode_id parentid;
	char *name;
	struct stat stat;
} fs_node;

// volume (per mount)
typedef struct fs_nspace {
	nspace_id id;
	fs_node root;						// fs_node for "/"
	struct netbfs_parms_t parms;
	struct bone_endpoint *ep;			// main connection
	struct bone_endpoint *nodemon_ep;	// node_monitor connection
	bool nodemon_notificating;
	sem_id sem;
	char *vol_name;
	
	char vol_decomp_temp[NETBFS_DATA_MAXSIZE];
	char vol_recv_buff[NETBFS_DATA_MAXSIZE];
	char vol_send_buff[sizeof(netbfs_cmd_t) + NETBFS_DATA_MAXSIZE + (NETBFS_DATA_MAXSIZE / 64) + 16 + 3];
	
	dir_cache_t *dir_cache;
} fs_nspace;


// cookie for fs_opendir, fs_readdir, fs_closedir, fs_free_dircookie, fs_rewinddir
typedef struct fs_dircookie {
	dir_cache_t *dir;
	entry_cache_t *current;
	entry_cache_t *entries;
} fs_dircookie;

// cookie for fs_open, fs_close, fs_read, fs_write, fs_free_cookie
typedef struct fs_file_cookie {
	struct bone_endpoint *ep;
	sem_id fileio_sem;
	int omode;
	char *fio_decomp_temp;
	char *fio_recv_buff;
	char *fio_send_buff;
} fs_file_cookie;


int32 api_version = B_CUR_FS_API_VERSION;
static bone_endpoint_info_t *mod_bone_ep = NULL;
static time_t mounted_time;

void dump(const char *cmd, uchar *data, uint32 size)
{
	int a = 0, i;
	uchar c;
	DPRINTF(1, (ID "  %4.4s +0 +1 +2 +3 +4 +5 +6 +7 +8 +9 +A +B +C +D +E +F  0123456789ABCDEF\n", cmd));
	DPRINTF(1, (ID "  ----------------------------------------------------  ----------------\n"));
	while(a < size) {
		DPRINTF(1, (ID "  %4.4X", a));
		for(i = 0; i < 16; i++) {
			if ((a + i) < size) {
				c = ((uchar *)data)[a + i];
				DPRINTF(1, (" %2.2X", c));
			} else {
				DPRINTF(1, ("   "));
			}
		}
		DPRINTF(1, ("  "));
		for(i = 0; i < 16; i++) {
			if ((a + i) < size) {
				c = ((uchar *)data)[a + i];
				DPRINTF(1, ("%c", isprint(c) ? c : '.'));
			}
		}
		a += 16;
		DPRINTF(1, ("\n"));
	}
}

#undef DEBUG
#define DEBUG 0
void clean_up(fs_nspace *vol)
{
	DPRINTF(DEBUG, (ID "clean_up(%p)\n", vol));
	if (vol->ep) {
		mod_bone_ep->close(vol->ep);
		mod_bone_ep->free(vol->ep);
	}
	if (vol->nodemon_ep) {
		if (vol->nodemon_notificating) mod_bone_ep->cancel_notification(vol->nodemon_ep, NOTIFY_RECV, vol);
		mod_bone_ep->close(vol->nodemon_ep);
		mod_bone_ep->free(vol->nodemon_ep);
	}
	if (vol->sem >= 0) {
		delete_sem(vol->sem);
	}
	c_make_empty_dir(vol->dir_cache);
	free(vol->root.name);
	free(vol->vol_name);
	free(vol);
	put_module(BONE_EP_MOD_NAME);
}

status_t lzo_compress(lzo_byte *in, lzo_uint in_len, lzo_byte *out, lzo_uint *out_len)
{
	lzo_byte *wrkmem = (lzo_bytep)malloc(LZO1X_1_MEM_COMPRESS);
	if (!wrkmem) return ENOMEM;
	if (lzo1x_1_compress(in, in_len, out, out_len, wrkmem) != LZO_E_OK) {
		free(wrkmem);
		return EIO;
	}
	free(wrkmem);
	return B_OK;
}

// wait for 'timeout' micro second until receive data comes.
ssize_t wait_recv(struct bone_endpoint *ep, bigtime_t timeout)
{
	ssize_t size, avail;
	bigtime_t ltime;
	ltime = system_time() + timeout;
	while ((system_time() < ltime) && (avail = mod_bone_ep->recv_avail(ep)) == 0) snooze(10);
	return avail;
}

// receive
#undef DEBUG
#define DEBUG 0
ssize_t receive(struct bone_endpoint *ep, void *data, size_t datalen)
{
	ssize_t size = mod_bone_ep->recv(ep, data, datalen, 0);
//	if (DEBUG) dump("recv", data, size);
	return size;
}

void clear_buff(struct bone_endpoint *ep)
{
	char dummy[4096];
	ssize_t size = 0;
	while (mod_bone_ep->recv_avail(ep) > 0) size += mod_bone_ep->recv(ep, dummy, sizeof(dummy), 0);
	DPRINTF(1, (ID "  ERROR RECOVERY. DROPPED %d BYTES.\n", size));
}

// receive reply data from server.
#undef DEBUG
#define DEBUG 1
status_t recv_reply(fs_nspace *vol, struct bone_endpoint *ep, replyinfo_t *replyinfo, iobuff_t *iobuff)
{
	netbfs_reply_t reply;
	char *_reply = (char *)&reply;
	ssize_t size;
	uint32 recvPos = 0;
	lzo_uint decomp_size;
	
	// read header
	while(recvPos < sizeof(netbfs_reply_t)) {
		if (wait_recv(ep, vol->parms.timeout) <= 0) return ETIMEDOUT;
		size = receive(ep, &_reply[recvPos], sizeof(netbfs_reply_t) - recvPos);
		if (size <= 0) { clear_buff(ep); DPRINTF(DEBUG, (ID "  EIO 1\n")); return EIO; }
		recvPos += size;
	}
	if ((reply.r_data_length > NETBFS_DATA_MAXSIZE) || (reply.r_header != NETBFS_REPLY_HEADER)) {
		clear_buff(ep);
		DPRINTF(DEBUG, (ID "  EIO 2\n"));
		return EIO;
	}
	
	// read data
	recvPos = 0;
	while(recvPos < reply.r_data_length) {
		if (wait_recv(ep, vol->parms.timeout) <= 0) return ETIMEDOUT;
		size = receive(ep, &iobuff->recv_buff[recvPos], reply.r_data_length - recvPos);
		if (size <= 0) { clear_buff(ep); DPRINTF(DEBUG, (ID "  EIO 3\n")); return EIO; }
		recvPos += size;
	}
	
	switch(reply.r_data_stat) {
		case NETBFS_UNCOMPRESSED: {
			replyinfo->data = iobuff->recv_buff;
			replyinfo->data_length = reply.r_data_length;
			break;
		}
		case NETBFS_LZO_COMPRESSED: {
			decomp_size = NETBFS_DATA_MAXSIZE;
			if (lzo1x_decompress_safe(iobuff->recv_buff, reply.r_data_length, iobuff->decomp_temp, &decomp_size, 0) != 0) {
				DPRINTF(DEBUG, (ID "  EIO 4\n"));
 				return EIO;
			}
			replyinfo->data = iobuff->decomp_temp;
			replyinfo->data_length = decomp_size;
			break;
		}
		default: {
			DPRINTF(DEBUG, (ID "  unknown compress type.\n"));
 			return EIO;
		}
	}
	
	replyinfo->status = reply.r_status;
	replyinfo->continuation = reply.r_continuation;
	
	return B_OK;
}

// send command to server.
#undef DEBUG
#define DEBUG 0
status_t send_cmd(fs_nspace *vol, struct bone_endpoint *ep, uint8 command,
					uint8 data_stat, uint32 data_length, void *data, iobuff_t *iobuff)
{
	lzo_byte *data_buff = (lzo_byte *)&iobuff->send_buff[sizeof(netbfs_cmd_t)];
	netbfs_cmd_t *cmd = (netbfs_cmd_t *)iobuff->send_buff;
	lzo_uint compressed_size;
	uint32 send_buff_size;
	ssize_t ssize;
	status_t err = B_OK;
	
	if (data_length > NETBFS_DATA_MAXSIZE) return EINVAL;
	
	if ((data_length == 0) || (data == NULL)) {
		data_length = 0;
		data_stat = NETBFS_UNCOMPRESSED;
	}
	
	switch(data_stat) {
		case NETBFS_UNCOMPRESSED:
			memcpy(data_buff, data, data_length);
			cmd->c_data_stat = NETBFS_UNCOMPRESSED;
			cmd->c_data_length = data_length;
			break;
		case NETBFS_LZO_COMPRESSED:
			compressed_size = NETBFS_DATA_MAXSIZE - sizeof(netbfs_cmd_t);
			if (lzo_compress((lzo_byte *)data, data_length, data_buff, &compressed_size) == 0) {
				if (compressed_size < data_length) {
					cmd->c_data_stat = NETBFS_LZO_COMPRESSED;
					cmd->c_data_length = compressed_size;
					break;
				}
			}
			// compress failed or incompressible
			memcpy(data_buff, data, data_length);
			cmd->c_data_stat = NETBFS_UNCOMPRESSED;
			cmd->c_data_length = data_length;
			break;
		default: return EINVAL;
	}
	
	cmd->c_header = NETBFS_COMMAND_HEADER;
	cmd->c_command = command;
	send_buff_size = sizeof(netbfs_cmd_t) + cmd->c_data_length;
	
	// send
	if ((ssize = mod_bone_ep->send(ep, iobuff->send_buff, send_buff_size, 0)) != send_buff_size) {
		DPRINTF(DEBUG, (ID "  send() error.\n"));
		err = EIO;
	} else {
//		if (DEBUG) dump("send", vol->send_buff, ssize);
	}
	
	return err;
}

// callback function for node_monitor connection
#undef DEBUG
#define DEBUG 0
void nodemon_callback(struct bone_endpoint *ep, int32 value, uint32 arg1, void *arg2 )
{
	fs_nspace *vol = (fs_nspace *)arg2;
	ssize_t size;
	
	DPRINTF(DEBUG, (ID "nodemon_callback(%p, %d, %d, %p)\n", ep, value, arg1, arg2));
	
	if (!vol) {
		DPRINTF(DEBUG, (ID "   Internal error ??? vol = NULL\n"));
		return;
	}
	
	// サーバーから変更情報を受信
//	_recv
//	notify_listener(B_ENTRY_REMOVED, vol->id, NETBFS_ROOTNODE_ID, 0, 2099888, NULL);
}

// connect to server
#undef DEBUG
#define DEBUG 0
int connect_to_server(fs_nspace *vol, struct bone_endpoint **ep_ptr)
{
	status_t err;
	int block;
	replyinfo_t replyinfo;
	iobuff_t iobuff;
	
	iobuff.decomp_temp = vol->vol_decomp_temp;
	iobuff.recv_buff = vol->vol_recv_buff;
	iobuff.send_buff = vol->vol_send_buff;
	
	vol->parms.server.sin_len = sizeof(vol->parms.server);
	DPRINTF(DEBUG, (ID "  connecting : id = \"%s\", sin_family = %d, "
				"s_addr = %d.%d.%d.%d, port = %d\n",
			vol->parms.idpass, vol->parms.server.sin_family,
			((uint8 *)&vol->parms.server.sin_addr.s_addr)[0],
			((uint8 *)&vol->parms.server.sin_addr.s_addr)[1],
			((uint8 *)&vol->parms.server.sin_addr.s_addr)[2],
			((uint8 *)&vol->parms.server.sin_addr.s_addr)[3],
			ntohs(vol->parms.server.sin_port)));
	
	// create bone_endpoint
	if ((err = mod_bone_ep->endpoint(ep_ptr, AF_INET, SOCK_STREAM, 0)) != B_OK) {
		DPRINTF(DEBUG, (ID "  can't get endpoint. %s\n", strerror(err)));
		return ENOSYS;
	}
	
	// 非ブロッキングに設定してサーバーへ接続
	//   blocking のままだと、応答の無いサーバーだとハングしてしまう。
	//   nonbloking にすることにより connect() は即リターンする。
	//   おそらく返り値は EINPROGRESS（接続中）になるが、あえて接続完了を待たずそのまま続行させ、
	//   実際、接続が成功したかどうかは recv() 時にわかる。
	block = 1;
	if (mod_bone_ep->setsockopt(*ep_ptr, SOL_SOCKET, SO_NONBLOCK, &block, sizeof(block)) < 0) {
		DPRINTF(DEBUG, (ID "  setsockopt(non-blocking mode) error\n"));
		err = B_ERROR;
		goto err_exit;
	}
	err = mod_bone_ep->connect(*ep_ptr, (struct sockaddr *)&vol->parms.server, sizeof(vol->parms.server));
	if ((err != B_OK) && (err != EISCONN) && (err != EINPROGRESS) && (err != EALREADY)) {
		DPRINTF(DEBUG, (ID "  connect error. %s\n", strerror(err)));
		err = ECONNREFUSED;
		goto err_exit;
	}
	
	// shift to blocking mode
	block = 0;
	if (mod_bone_ep->setsockopt(*ep_ptr, SOL_SOCKET, SO_NONBLOCK, &block, sizeof(block)) < 0) {
		DPRINTF(DEBUG, (ID "  setsockopt(blocking mode) error\n"));
		err = EIO;
		goto err_exit;
	}
	
	// receive prompt
	if ((err = recv_reply(vol, *ep_ptr, &replyinfo, &iobuff)) != B_OK) {
		DPRINTF(DEBUG, (ID "  recv_reply() error: %s\n", strerror(err)));
		goto err_exit;
	}
	if (replyinfo.status != NETBFS_OK) {
		DPRINTF(DEBUG, (ID "  server invalid response. (login)\n"));
		err = EIO;
		goto err_exit;
	}
	
	// send "ID@PASS@SHARE"
	if ((err = send_cmd(vol, *ep_ptr, NETBFS_CMD_LOGIN, NETBFS_UNCOMPRESSED, strlen(vol->parms.idpass), vol->parms.idpass, &iobuff)) != B_OK) {
		DPRINTF(DEBUG, (ID "  send_cmd() error: %s\n", strerror(err)));
		goto err_exit;
	}
	
	// receive authorization result
	if ((err = recv_reply(vol, *ep_ptr, &replyinfo, &iobuff)) != B_OK) {
		DPRINTF(DEBUG, (ID "  recv_reply() error: %s\n", strerror(err)));
		goto err_exit;
	}
	switch(replyinfo.status) {
		case NETBFS_OK:
			DPRINTF(DEBUG, (ID "  login successful!\n"));
			break;
		case NETBFS_EACCES:
			DPRINTF(DEBUG, (ID "  login incorrect.\n"));
			err = EACCES;
			goto err_exit;
		default:
			DPRINTF(DEBUG, (ID "  server invalid response. (login auth)\n"));
			err = EIO;
			goto err_exit;
	}
	
	return 0;
	
err_exit:
	mod_bone_ep->close(*ep_ptr);
	mod_bone_ep->free(*ep_ptr);
	*ep_ptr = NULL;
	return err;
}

// mount
#undef DEBUG
#define DEBUG 0
int fs_mount(nspace_id nsid, const char *devname, ulong flags,
			netbfs_parms_t *parms, size_t len, fs_nspace **data, vnode_id *vnid)
{
	fs_nspace *vol;
	int timeout, err, volnameip_len;
	sem_id sem;
	char volnameip[17];
	char *share_name;
	
	DPRINTF(DEBUG, (ID "fs_mount(%d, %s, %d, %p, %d, %p, %p)\n", nsid, devname, flags, parms, len, data, vnid));
	
	if ((parms == NULL) || (len < sizeof(netbfs_parms_t))) {
		DPRINTF(DEBUG, (ID "  invalid parms.\n"));
		return EINVAL;
	}
	
	if ((err = get_module(BONE_EP_MOD_NAME, (module_info **)&mod_bone_ep)) != B_OK) {
		DPRINTF(DEBUG, (ID "  get_module(%s) failed. %s\n", BONE_EP_MOD_NAME, strerror(err)));
		return ENOSYS;
	}
	
	//
	vol = (fs_nspace *)calloc(sizeof(fs_nspace), 1);
	if (vol == NULL) {
		put_module(BONE_EP_MOD_NAME);
		return ENOMEM;
	}
	*data = vol;
	vol->sem = B_BAD_VALUE;
	vol->id = nsid;
	vol->root.id = NETBFS_ROOTNODE_ID;
	*vnid = vol->root.id;
	vol->parms = *parms;
	
	if ((vol->root.name = (char *)malloc(2)) == NULL) {
		clean_up(vol);
		return ENOMEM;
	}
	strcpy(vol->root.name, "/");
	
	if ((sem = create_sem(1, "netbfs_bone_sem")) < B_OK) {
		DPRINTF(DEBUG, (ID "  create_sem() failed. %s\n", strerror(sem)));
		clean_up(vol);
		return sem;
	}
	vol->sem = sem;
	
	// set volume name
	sprintf(volnameip, "%d.%d.%d.%d_",
			((uint8 *)&vol->parms.server.sin_addr.s_addr)[0],
			((uint8 *)&vol->parms.server.sin_addr.s_addr)[1],
			((uint8 *)&vol->parms.server.sin_addr.s_addr)[2],
			((uint8 *)&vol->parms.server.sin_addr.s_addr)[3]);
	volnameip_len = strlen(volnameip);
	if ((share_name = strchr(vol->parms.idpass, '@')) == NULL) {
		clean_up(vol);
		return EINVAL;
	}
	if ((share_name = strchr(share_name + 1, '@')) == NULL) {
		clean_up(vol);
		return EINVAL;
	}
	share_name++;
	if ((vol->vol_name = (char *)malloc(volnameip_len + strlen(share_name) + 1)) == NULL) {
		clean_up(vol);
		return ENOMEM;
	}
	strcpy(vol->vol_name, volnameip);
	strcpy(&vol->vol_name[volnameip_len], share_name);
	DPRINTF(DEBUG, (ID "  volume name = %s\n", vol->vol_name));
	
	// connect (main)
	DPRINTF(DEBUG, (ID "  Connecting (main).\n"));
	if ((err = connect_to_server(vol, &vol->ep)) != 0) {
		DPRINTF(DEBUG, (ID "  main connect_to_server() error: %s\n", strerror(err)));
		clean_up(vol);
		return err;
	}
	
	// connect (node_monitor)
	DPRINTF(DEBUG, (ID "  Connecting (node monitor).\n"));
	if ((err = connect_to_server(vol, &vol->nodemon_ep)) != 0) {
		DPRINTF(DEBUG, (ID "  node_monitor connect_to_server() error: %s\n", strerror(err)));
		clean_up(vol);
		return err;
	}
	
	// initialization for node_monitor connection
	if ((err = mod_bone_ep->request_notification(vol->nodemon_ep, NOTIFY_RECV, false, nodemon_callback, 0, vol)) != B_OK) {
		DPRINTF(DEBUG, (ID "  request_notification() error: %s\n", strerror(err)));
		clean_up(vol);
		return err;
	}
	vol->nodemon_notificating = true;
	//argv[0] = "nodemon";
	//send_cmd(vol->nodemon_ep, 1, argv);
	
	// create "/" vnode
	if ((err = new_vnode(nsid, *vnid, &vol->root)) != B_OK) {
		DPRINTF(DEBUG, (ID "  new_vnode() failed. %s\n", strerror(err)));
		clean_up(vol);
		return EINVAL;
	}
	
	DPRINTF(DEBUG, (ID "  nsid = %d , data = %p\n", nsid, data));
	
	time(&mounted_time);
	
	return B_OK;
}

// unmount
#undef DEBUG
#define DEBUG 0
int fs_unmount(fs_nspace *vol)
{
	DPRINTF(DEBUG, (ID "fs_unmount(%p)\n", vol));
	clean_up(vol);
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_rfsstat(fs_nspace *vol, struct fs_info *info)
{
	DPRINTF(DEBUG, (ID "fs_rfsstat(%p, %p)\n", vol, info));
	
//	info->dev = vol->fd;
//	info->root = vol->root.id;
//	info->flags = B_FS_HAS_MIME | B_FS_HAS_ATTR | B_FS_HAS_QUERY | B_FS_IS_PERSISTENT | B_FS_IS_READONLY;
//	info->flags = B_FS_HAS_MIME | B_FS_HAS_ATTR | B_FS_HAS_QUERY | B_FS_IS_SHARED | B_FS_IS_READONLY;
	info->flags = B_FS_IS_SHARED | B_FS_IS_READONLY;
	info->block_size = 4096;
	info->io_size = 65536;
	info->total_blocks = 0;
	info->free_blocks = 0;
	strcpy(info->device_name, "");
	strcpy(info->volume_name, vol->vol_name);
	strcpy(info->fsh_name, "netbfs");
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_walk(fs_nspace *vol, fs_node *baseNode, const char *file, char **newpath, vnode_id *vnid)
{
	void *cookie;
	fs_node *newNode = NULL;
	dir_cache_t *dir;
	entry_cache_t *entry;
	int baselen = strlen(baseNode->name);
	char path[baselen + strlen(file) + 2];	// + 2 は、"/" と \0 のバイト数
	char *_path = path;
	
	DPRINTF(DEBUG, (ID "fs_walk(%p, %p, \"%s\", *, *)  basedir = %s\n", vol, baseNode, file, baseNode->name));
	
	if (strcmp(file, ".") == 0) {
		// current directory
		DPRINTF(DEBUG, (ID "  \".\" requested. vnid = %d\n", baseNode->id));
		*vnid = baseNode->id;
		if (get_vnode(vol->id, *vnid, (void **)&newNode) != 0) return EINVAL;
		return B_OK;
	}
	if (strcmp(file, "..") == 0) {
		// parent directory
		DPRINTF(DEBUG, (ID "  \"..\" requested. vnid = %d\n", baseNode->parentid));
		*vnid = baseNode->parentid;
		if (get_vnode(vol->id, *vnid, (void **)&newNode) != 0) return EINVAL;
		return B_OK;
	}
	
	// look up file in the directory from dir_cache
	if ((dir = c_find_dir(vol->dir_cache, baseNode->id)) == NULL) {
		// not exists in cache
		DPRINTF(DEBUG, (ID "  cache not found.\n"));
		
		// force read from server
		if (fs_opendir(vol, baseNode, &cookie) != 0) return ENOENT;
		fs_closedir(vol, baseNode, cookie);
		fs_free_dircookie(vol, baseNode, cookie);
		
		if ((dir = c_find_dir(vol->dir_cache, baseNode->id)) == NULL) {
			// not exists in cache !?
			DPRINTF(DEBUG, (ID "  ?? dir not found.\n"));
			return ENOENT;
		}
	}
	if ((entry = c_find_entry_name(dir, file)) == NULL) return ENOENT;
	
	*vnid = entry->stat.st_ino;
	if (get_vnode(vol->id, *vnid, (void **)&newNode) != 0) {
		DPRINTF(DEBUG, (ID "  get_vnode() failed\n"));
		return EINVAL;
	}
	DPRINTF(DEBUG, (ID "  fs_walk newNode = %p\n", newNode));
	
	// make fullpath name
	if (strcmp(baseNode->name, "/") == 0) {
		strcpy(_path++, "/");
		strcpy(_path, file);
	} else {
		strcpy(_path, baseNode->name);
		_path += baselen;
		strcpy(_path++, "/");
		strcpy(_path, file);
	}
	DPRINTF(DEBUG, (ID "  path = %s\n", path));
	
	// 
	newNode->parentid = baseNode->id;
	newNode->name = (char *)malloc(strlen(path) + 1);
	strcpy(newNode->name, path);
	newNode->stat.st_ino = entry->stat.st_ino;
	newNode->stat.st_mode = entry->stat.st_mode;
	newNode->stat.st_size = entry->stat.st_size;
	newNode->stat.st_blksize = entry->stat.st_blksize;
	newNode->stat.st_atime = entry->stat.st_atime;
	newNode->stat.st_mtime = entry->stat.st_mtime;
	newNode->stat.st_ctime = entry->stat.st_ctime;
	newNode->stat.st_crtime = entry->stat.st_crtime;
	
	return 0;
}

#undef DEBUG
#define DEBUG 0
int fs_read_vnode(fs_nspace *vol, vnode_id vnid, char r, fs_node **node)
{
	fs_node *newNode;
	
	DPRINTF(DEBUG, (ID "fs_read_vnode(%p, %Ld, %d, *)\n", vol, vnid, r));
	
	newNode = (fs_node *)calloc(sizeof(fs_node), 1);
	if (newNode == NULL) return ENOMEM;
	
	if (vnid == NETBFS_ROOTNODE_ID) {
		DPRINTF(DEBUG, (ID "   fs_read_vnode - root node requested.\n"));
		memcpy(newNode, &vol->root, sizeof(fs_node));
		*node = newNode;
	} else {
		newNode->id = vnid;
		*node = newNode;
	}
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_write_vnode(fs_nspace *vol, fs_node *node, char r)
{
	DPRINTF(DEBUG, (ID "fs_write_vnode(%p, %p, %d)\n", vol, node, r));
	
	if (node->id != NETBFS_ROOTNODE_ID) {
		if (node) {
			free(node->name);
			free(node);
		}
	}
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_rstat(fs_nspace *vol, fs_node *node, struct stat *st)
{
	DPRINTF(DEBUG, (ID "fs_rstat(%p, %p, %p)\n  node->name = %s", vol, node, st, node->name));
	
	switch(node->id) {
		case NETBFS_ROOTNODE_ID:
			st->st_dev = vol->id;
			st->st_ino = node->id;
			st->st_mode = S_IFDIR | 0555;
			st->st_nlink = 1;
			st->st_uid = 0;
			st->st_gid = 0;
		    st->st_size = 0;
		    st->st_blksize = 65536;
		    st->st_atime = mounted_time;
			st->st_mtime = mounted_time;
			st->st_ctime = mounted_time;
			st->st_crtime = mounted_time;
			break;
		default:
			st->st_dev = vol->id;
			st->st_ino = node->id;
			st->st_mode = node->stat.st_mode;
			st->st_nlink = 1;
			st->st_uid = 0;
			st->st_gid = 0;
		    st->st_size = node->stat.st_size;
		    st->st_blksize = node->stat.st_blksize;
		    st->st_atime = node->stat.st_atime;
			st->st_mtime = node->stat.st_mtime;
			st->st_ctime = node->stat.st_ctime;
			st->st_crtime = node->stat.st_crtime;
			break;
//		default:
//			DPRINTF(DEBUG, (ID "fs_rstat B_ERROR\n"));
//			return B_ERROR;
	}
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_open(fs_nspace *vol, fs_node *node, int omode, fs_file_cookie **cookie)
{
	fs_file_cookie *file_cookie;
	
	if (node->id != NETBFS_ROOTNODE_ID)
//		DPRINTF(DEBUG, (ID "fs_open(%p, %p, %d, *)  node->name = %s\n", vol, node, omode, node->name));
		DPRINTF(DEBUG, (ID "fs_open(%p, %p, %d, *)\n", vol, node, omode));
	
	switch(node->id) {
		case NETBFS_ROOTNODE_ID:
			*cookie = NULL;
			break;
		default: {
			char wk[256];
			sprintf(wk, "netbfs_fileio_sem_%lld", node->id);
			if ((file_cookie = (fs_file_cookie *)calloc(sizeof(fs_file_cookie), 1)) == NULL) {
				return ENOMEM;
			}
			file_cookie->omode = omode;
			file_cookie->fileio_sem = create_sem(1, wk);
			*cookie = file_cookie;
		}
	}
	
	if (node->id != NETBFS_ROOTNODE_ID)
		DPRINTF(DEBUG, (ID "  cookie = %p\n", *cookie));
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_read(fs_nspace *vol, fs_node *node, fs_file_cookie *cookie, off_t pos, void *buf, size_t *len)
{
	int err = B_OK, namelen = strlen(node->name);
	char buff[sizeof(fopen_t) + namelen];
	replyinfo_t replyinfo;
	fio_t fio;
	char *bufptr = buf, *buflimit = &((char *)buf)[*len];
	uint32 rlen;
	iobuff_t iobuff;
	
	DPRINTF(DEBUG, (ID "fs_read(%p, %p, %p, %d,", vol, node, cookie, pos));
//	DPRINTF(DEBUG, (" %p, %d)  node->name = %s\n", buf, *len, node->name));
	DPRINTF(DEBUG, (" %p, %d)\n", buf, *len));
	
	if (cookie == NULL) return EINVAL;
	
	iobuff.decomp_temp = cookie->fio_decomp_temp;
	iobuff.recv_buff = cookie->fio_recv_buff;
	iobuff.send_buff = cookie->fio_send_buff;
	
//	DPRINTF(DEBUG, (ID "  1 %p\n", cookie));
	
	// enter critical section
	if (acquire_sem(cookie->fileio_sem) != B_OK) {
		DPRINTF(DEBUG, (ID "  acquire_sem() failed\n"));
		return ETIMEDOUT;
	}
	
//	DPRINTF(DEBUG, (ID "  2 %p\n", cookie));
	
	// connect, if first access
	if (cookie->ep == NULL) {
		
//		DPRINTF(DEBUG, (ID "  first access.\n"));
		
		if (connect_to_server(vol, &cookie->ep) != 0) {
			DPRINTF(DEBUG, (ID "  fs_read() connect failed.\n"));
			err = EIO;
			goto exit;
		}
		
		cookie->fio_decomp_temp = (char *)malloc(NETBFS_DATA_MAXSIZE);
		cookie->fio_recv_buff = (char *)malloc(NETBFS_DATA_MAXSIZE);
		cookie->fio_send_buff = (char *)malloc(sizeof(netbfs_cmd_t) + NETBFS_DATA_MAXSIZE + (NETBFS_DATA_MAXSIZE / 64) + 16 + 3);
		
		if (!cookie->fio_decomp_temp || !cookie->fio_recv_buff || !cookie->fio_send_buff) {
			free(cookie->fio_decomp_temp);
			free(cookie->fio_recv_buff);
			free(cookie->fio_send_buff);
			err = ENOMEM;
			goto exit;
		}
		
		iobuff.decomp_temp = cookie->fio_decomp_temp;
		iobuff.recv_buff = cookie->fio_recv_buff;
		iobuff.send_buff = cookie->fio_send_buff;
		
//		DPRINTF(DEBUG, (ID "  3 %p\n", cookie));
		
		// send NETBFS_CMD_FOPEN
		((fopen_t *)buff)->fo_omode = cookie->omode;
		((fopen_t *)buff)->fo_namelen = namelen;
		memcpy(&buff[sizeof(fopen_t)], node->name, namelen);
		if ((err = send_cmd(vol, cookie->ep, NETBFS_CMD_FOPEN, NETBFS_UNCOMPRESSED, namelen + sizeof(fopen_t), buff, &iobuff)) != B_OK) {
			DPRINTF(DEBUG, (ID "  send_cmd() error: %s\n", strerror(err)));
			goto exit;
		}
		
//		DPRINTF(DEBUG, (ID "  4 %p\n", cookie));
		
		// receive reply
		if ((err = recv_reply(vol, cookie->ep, &replyinfo, &iobuff)) != B_OK) {
			DPRINTF(DEBUG, (ID "  recv_reply() error 1: %s\n", strerror(err)));
			goto exit;
		}
//		DPRINTF(DEBUG, (ID "  5 %p\n", cookie));
		if (replyinfo.status != NETBFS_OK) {
			DPRINTF(DEBUG, (ID "  recv_reply() error 2: %s\n", strerror(err)));
			err = EIO;
			goto exit;
		}
//		DPRINTF(DEBUG, (ID "  6 %p\n", cookie));
	}
	
	// send NETBFS_CMD_FREAD
//	DPRINTF(DEBUG, (ID "  send NETBFS_CMD_FREAD\n"));
	fio.fio_pos = pos;
	fio.fio_len = *len;
	if ((err = send_cmd(vol, cookie->ep, NETBFS_CMD_FREAD, NETBFS_UNCOMPRESSED, sizeof(fio_t), &fio, &iobuff)) != B_OK) {
		DPRINTF(DEBUG, (ID "  send_cmd() error: %s\n", strerror(err)));
		goto exit;
	}
	
//	DPRINTF(DEBUG, (ID "  8 %p\n", cookie));
	// read data
	rlen = 0;
	do {
//		DPRINTF(DEBUG, (ID "  rlen = %d\n", rlen));
//		DPRINTF(DEBUG, (ID "  9 %p\n", cookie));
		if ((err = recv_reply(vol, cookie->ep, &replyinfo, &iobuff)) != B_OK) {
			DPRINTF(DEBUG, (ID "  recv_reply() error 3: %s\n", strerror(err)));
			goto exit;
		}
//		DPRINTF(DEBUG, (ID "  10 %p\n", cookie));
		if (replyinfo.status != NETBFS_OK) {
			DPRINTF(DEBUG, (ID "  recv_reply() error 4: %s\n", strerror(err)));
			err = EIO;
			goto exit;
		}
//		DPRINTF(DEBUG, (ID "  11 %p\n", cookie));
		memcpy(bufptr, replyinfo.data, min_c(replyinfo.data_length, buflimit - bufptr));
		bufptr += replyinfo.data_length;
		rlen += replyinfo.data_length;
//		DPRINTF(DEBUG, (ID "  12 %p\n", cookie));
	} while (replyinfo.continuation);
	
	
exit:	
//	DPRINTF(DEBUG, (ID "  13 %p\n", cookie));
	// leave critical section
	release_sem(cookie->fileio_sem);
	*len = rlen;
	DPRINTF(DEBUG, (ID "  14 %p\n", cookie));
	
	return err;
}

#undef DEBUG
#define DEBUG 0
int fs_write(fs_nspace *vol, fs_node *node, fs_file_cookie *cookie, off_t pos, const void *buf, size_t *len)
{
	DPRINTF(DEBUG, (ID "fs_write()\n"));
	
	if (cookie == NULL) return EINVAL;
	
	return EIO;
/*	
	if (cookie->ep == NULL) {
		// 最初のアクセスで fopen する
		
	}
	
	// 
	
	
	return B_OK;
*/
}

#undef DEBUG
#define DEBUG 0
int fs_close(fs_nspace *vol, fs_node *node, fs_file_cookie *cookie)
{
	if (node->id != NETBFS_ROOTNODE_ID)
		DPRINTF(DEBUG, (ID "fs_close(%p, %p, %p)\n", vol, node, cookie));
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_free_cookie(fs_nspace *vol, fs_node *node, fs_file_cookie *cookie)
{
	if (node->id != NETBFS_ROOTNODE_ID)
		DPRINTF(DEBUG, (ID "fs_free_cookie(%p, %p, *)\n", vol, node));
	
	if (cookie) {
		if (cookie->ep) {
			mod_bone_ep->close(cookie->ep);
			mod_bone_ep->free(cookie->ep);
			free(cookie->fio_decomp_temp);
			free(cookie->fio_recv_buff);
			free(cookie->fio_send_buff);
		}
		delete_sem(cookie->fileio_sem);
		free(cookie);
	}
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_access(void *ns, void *node, int mode)
{
	DPRINTF(DEBUG, (ID "fs_access(%p, %p, %d)\n", ns, node, mode));
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_create(fs_nspace *vol, fs_node *dir, const char *name, int omode,
	int perms, vnode_id *vnid, fs_file_cookie **cookie)
{
	DPRINTF(DEBUG, (ID "fs_create()\n"));
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_mkdir(fs_nspace *vol, fs_node *dir, const char *name, int perms)
{
	DPRINTF(DEBUG, (ID "fs_mkdir()\n"));
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_symlink(fs_nspace *vol, fs_node *dir, const char *name, const char *path)
{
	DPRINTF(DEBUG, (ID "fs_symlink()\n"));
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_rename(fs_nspace *vol, fs_node *olddir, const char *oldname, fs_node *newdir, const char *newname)
{
	DPRINTF(DEBUG, (ID "fs_rename()\n"));
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_unlink(fs_nspace *vol, fs_node *dir, const char *name)
{
	DPRINTF(DEBUG, (ID "fs_unlink()\n"));
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_rmdir(fs_nspace *vol, fs_node *dir, const char *name)
{
	DPRINTF(DEBUG, (ID "fs_rmdir()\n"));
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_readlink(fs_nspace *vol, fs_node *node, char *buf, size_t *bufsize)
{
	DPRINTF(DEBUG, (ID "fs_readlink()\n"));
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_opendir(fs_nspace *vol, fs_node *node, fs_dircookie **cookie)
{
	status_t err;
	replyinfo_t replyinfo;
	fs_dircookie *dir_cookie;
	char *dataptr, *lastptr;
	uint32 numBytes;
	fstat_t *fstat;
	struct stat stat;
	entry_cache_t *entry;
	iobuff_t iobuff;
	
	iobuff.decomp_temp = vol->vol_decomp_temp;
	iobuff.recv_buff = vol->vol_recv_buff;
	iobuff.send_buff = vol->vol_send_buff;
	
	DPRINTF(DEBUG, (ID "fs_opendir(%p, %p, %p) dir = %s\n", vol, node, cookie, node->name));
	
	if (!vol || !node || !cookie) return EINVAL;
	if (!node->name) return EINVAL;
	
	// create dircookie
	if ((dir_cookie = (fs_dircookie *)malloc(sizeof(fs_dircookie))) == NULL) return ENOMEM;
	
	// search cache
	if ((dir_cookie->dir = c_find_dir(vol->dir_cache, node->id)) != NULL) {
		// found in cache
		DPRINTF(DEBUG, (ID "  found in cache.\n"));
	} else {
		// not found in cache
		DPRINTF(DEBUG, (ID "  not found in cache.\n"));
		
		// enter critical section
		if (acquire_sem(vol->sem) != B_OK) {
			DPRINTF(DEBUG, (ID "  acquire_sem() failed\n"));
			free(dir_cookie);
			return ETIMEDOUT;
		}
		
		// send NETBFS_CMD_READDIR
		if ((err = send_cmd(vol, vol->ep, NETBFS_CMD_READDIR, NETBFS_UNCOMPRESSED, strlen(node->name), node->name, &iobuff)) != B_OK) {
			DPRINTF(DEBUG, (ID "  send_cmd() error: %s\n", strerror(err)));
			goto err_exit;
		}
		
		// create new cache
		dir_cookie->dir = c_add_dir(&vol->dir_cache, node->name, node->id);
		
		// receive entry data
		do {
			if ((err = recv_reply(vol, vol->ep, &replyinfo, &iobuff)) != B_OK) {
				DPRINTF(DEBUG, (ID "  recv_reply() error: %s\n", strerror(err)));
				goto err_exit;
			}
			if (replyinfo.status != NETBFS_OK) {
				DPRINTF(DEBUG, (ID "  recv_reply() error status: %s\n", strerror(replyinfo.status)));
				err = EIO;
				goto err_exit;
			}
			dataptr = replyinfo.data;
			lastptr = &dataptr[replyinfo.data_length];
			while (dataptr < lastptr) {
				fstat = (fstat_t *)dataptr;
				stat.st_ino = fstat->fs_ino;
				stat.st_mode = fstat->fs_mode;
				stat.st_size = fstat->fs_size;
				stat.st_atime = fstat->fs_atime;
				stat.st_mtime = fstat->fs_mtime;
				stat.st_ctime = fstat->fs_ctime;
				stat.st_crtime = fstat->fs_crtime;
				dataptr += sizeof(fstat_t);
				entry = c_add_entry(dir_cookie->dir, dataptr, fstat->fs_namelen, &stat);
				dataptr += fstat->fs_namelen;
			}
		} while(replyinfo.continuation);
		
		// leave critical section
		release_sem(vol->sem);
	}
	
	dir_cookie->entries = dir_cookie->dir->entries;
	while (dir_cookie->entries->next) dir_cookie->entries = dir_cookie->entries->next;
	dir_cookie->current = dir_cookie->entries;
	
	*cookie = dir_cookie;
	return 0;
	
err_exit:
	c_remove_dir(&vol->dir_cache, node->id);
	free(dir_cookie);
	release_sem(vol->sem);
	return err;
}

#undef DEBUG
#define DEBUG 0
int fs_closedir(fs_nspace *vol, fs_node *node, fs_dircookie *cookie)
{
	DPRINTF(DEBUG, (ID "fs_closedir(%p, %p, %p) dir = %s\n", vol, node, cookie, node->name));
	if (!vol || !node || !cookie) return EINVAL;
	return 0;
}

#undef DEBUG
#define DEBUG 0
int fs_free_dircookie(fs_nspace *vol, fs_node *node, fs_dircookie *cookie)
{
	DPRINTF(DEBUG, (ID "fs_free_dircookie(%p, %p, %p) dir = %s\n", vol, node, cookie, node->name));
	if (!vol || !node || !cookie) return EINVAL;
	free(cookie);
	return 0;
}

#undef DEBUG
#define DEBUG 0
int fs_rewinddir(fs_nspace *vol, fs_node *node, fs_dircookie *cookie)
{
	DPRINTF(DEBUG, (ID "fs_rewinddir(%p, %p, %p) dir = %s\n", vol, node, cookie, node->name));
	if (!vol || !node || !cookie) return EINVAL;
	cookie->current = cookie->entries;
	return 0;
}

#undef DEBUG
#define DEBUG 0
int fs_readdir(fs_nspace *vol, fs_node *node, fs_dircookie *cookie,
	long *num, struct dirent *entry, size_t bufsize)
{
	DPRINTF(DEBUG, (ID "fs_readdir(%p, %p, %p, *, %p, %d) dir = %s\n", vol, node, cookie, entry, bufsize, node->name));
	
	if (!vol || !node || !cookie || !num || !entry) return EINVAL;
	
	if (cookie->current) {
		*num = 1;
		entry->d_dev = vol->id;
		entry->d_ino = cookie->current->stat.st_ino;
		entry->d_reclen = 2 * (sizeof(dev_t) + sizeof(ino_t)) + sizeof(unsigned short)
							+ strlen(cookie->current->leafname) + 1;
		strcpy(entry->d_name, cookie->current->leafname);
//		DPRINTF(DEBUG, (ID "  name = %s\n", cookie->current->leafname));
		cookie->current = cookie->current->prev;
	} else {
		*num = 0;
	}
	return 0;
}

#undef DEBUG
#define DEBUG 0
int fs_wstat(fs_nspace *vol, fs_node *node, struct stat *st, long mask)
{
	DPRINTF(DEBUG, (ID "fs_wstat(%p, %p, %p, %d)\n", vol, node, st, mask));
	
	if (!vol || !node || !st) return EINVAL;
	
	return B_OK;
}

#undef DEBUG
#define DEBUG 0
int fs_wfsstat(fs_nspace *vol, struct fs_info *info, long mask)
{
	DPRINTF(DEBUG, (ID "fs_wfsstat(%p, %p, %d)\n", vol, info, mask));
	
	if (!vol || !info) return EINVAL;
	
	return B_OK;
}

// ----
#undef DEBUG
#define DEBUG 1
int fs_remove_vnode(void *ns, void *node, char r)
{
	DPRINTF(DEBUG, (ID "fs_remove_vnode\n"));
	return B_ERROR;
}

int fs_secure_vnode(void *ns, void *node)
{
	DPRINTF(DEBUG, (ID "fs_secure_vnode\n"));
	return B_ERROR;
}

int fs_link(void *ns, void *dir, const char *name, void *node)
{
	DPRINTF(DEBUG, (ID "fs_link\n"));
	return B_ERROR;
}

int fs_readv(void *ns, void *node, void *cookie, off_t pos, const iovec *vec, size_t count, size_t *len)
{
	DPRINTF(DEBUG, (ID "fs_readv\n"));
	return B_ERROR;
}

int fs_writev(void *ns, void *node, void *cookie, off_t pos, const iovec *vec, size_t count, size_t *len)
{
	DPRINTF(DEBUG, (ID "fs_writev\n"));
	return B_ERROR;
}

int fs_ioctl(void *ns, void *node, void *cookie, int cmd, void *buf, size_t len)

{
	DPRINTF(DEBUG, (ID "fs_ioctl\n"));
	return B_ERROR;
}

int fs_setflags(void *ns, void *node, void *cookie, int flags)
{
	DPRINTF(DEBUG, (ID "fs_setflags\n"));
	return B_ERROR;
}

int fs_fsync(void *ns, void *node)
{
	DPRINTF(DEBUG, (ID "fs_fsync\n"));
	return B_ERROR;
}

int fs_initialize(const char *devname, void *parms, size_t len)
{
	DPRINTF(DEBUG, (ID "fs_initialize\n"));
	return B_ERROR;
}

int fs_sync(void *ns)
{
	DPRINTF(DEBUG, (ID "fs_sync\n"));
	return B_ERROR;
}

int fs_select(void *ns, void *node, void *cookie, uint8 event, uint32 ref, selectsync *sync)
{
	DPRINTF(DEBUG, (ID "fs_select\n"));
	return B_ERROR;
}

int fs_deselect(void *ns, void *node, void *cookie, uint8 event, selectsync *sync)
{
	DPRINTF(DEBUG, (ID "fs_deselect\n"));
	return B_ERROR;
}

int fs_open_indexdir(void *ns, void **cookie)
{
	DPRINTF(DEBUG, (ID "fs_open_indexdir\n"));
	return B_ERROR;
}

int fs_close_indexdir(void *ns, void *cookie)
{
	DPRINTF(DEBUG, (ID "fs_close_indexdir\n"));
	return B_ERROR;
}

int fs_free_indexdircookie(void *ns, void *node, void *cookie)
{
	DPRINTF(DEBUG, (ID "fs_free_indexdircookie\n"));
	return B_ERROR;
}

int fs_rewind_indexdir(void *ns, void *cookie)
{
	DPRINTF(DEBUG, (ID "fs_rewind_indexdir\n"));
	return B_ERROR;
}

int fs_read_indexdir(void *ns, void *cookie, long *num, struct dirent *buf, size_t bufsize)
{
	DPRINTF(DEBUG, (ID "fs_read_indexdir\n"));
	return B_ERROR;
}

int fs_create_index(void *ns, const char *name, int type, int flags)
{
	DPRINTF(DEBUG, (ID "fs_create_index\n"));
	return B_ERROR;
}

int fs_remove_index(void *ns, const char *name)
{
	DPRINTF(DEBUG, (ID "fs_remove_index\n"));
	return B_ERROR;
}

int fs_rename_index(void *ns, const char *oldname, const char *newname)
{
	DPRINTF(DEBUG, (ID "fs_rename_index\n"));
	return B_ERROR;
}

int fs_stat_index(void *ns, const char *name, struct index_info *buf)
{
	DPRINTF(DEBUG, (ID "fs_stat_index\n"));
	return B_ERROR;
}

int fs_open_attrdir(void *ns, void *node, void **cookie)
{
	DPRINTF(DEBUG, (ID "fs_open_attrdir\n"));
	return B_ERROR;
}

int fs_close_attrdir(void *ns, void *node, void *cookie)
{
	DPRINTF(DEBUG, (ID "fs_close_attrdir\n"));
	return B_ERROR;
}

int fs_free_attrdircookie(void *ns, void *node, void *cookie)
{
	DPRINTF(DEBUG, (ID "fs_free_attrdircookie\n"));
	return B_ERROR;
}

int fs_rewind_attrdir(void *ns, void *node, void *cookie)
{
	DPRINTF(DEBUG, (ID "fs_rewind_attrdir\n"));
	return B_ERROR;
}

int fs_read_attrdir(void *ns, void *node, void *cookie, long *num, struct dirent *buf, size_t bufsize)
{
	DPRINTF(DEBUG, (ID "fs_read_attrdir\n"));
	return B_ERROR;
}

int fs_writeattr(void *ns, void *node, const char *name, int type, const void *buf, size_t *len, off_t pos)
{
	DPRINTF(DEBUG, (ID "fs_writeattr\n"));
	return B_ERROR;
}

int fs_readattr(void *ns, void *node, const char *name, int type, void *buf, size_t *len, off_t pos)
{
	DPRINTF(DEBUG, (ID "fs_readattr\n"));
	return B_ERROR;
}

int fs_removeattr(void *ns, void *node, const char *name)
{
	DPRINTF(DEBUG, (ID "fs_removeattr\n"));
	return B_ERROR;
}

int fs_rename_attr(void *ns, void *node, const char *oldname, const char *newname)
{
	DPRINTF(DEBUG, (ID "fs_rename_attr\n"));
	return B_ERROR;
}

int fs_statattr(void *ns, void *node, const char *name, struct attr_info *buf)
{
	DPRINTF(DEBUG, (ID "fs_statattr\n"));
	return B_ERROR;
}

int fs_open_query(void *ns, const char *query, ulong flags, port_id port, long token, void **cookie)
{
	DPRINTF(DEBUG, (ID "fs_open_query\n"));
	return B_ERROR;
}

int fs_close_query(void *ns, void *cookie)
{
	DPRINTF(DEBUG, (ID "fs_close_query\n"));
	return B_ERROR;
}

int fs_free_querycookie(void *ns, void *node, void *cookie)
{
	DPRINTF(DEBUG, (ID "fs_free_querycookie\n"));
	return B_ERROR;
}

int fs_read_query(void *ns, void *cookie, long *num, struct dirent *buf, size_t bufsize)
{
	DPRINTF(DEBUG, (ID "fs_read_query\n"));
	return B_ERROR;
}

//
vnode_ops fs_entry =
{
	(op_read_vnode *)&fs_read_vnode,
	(op_write_vnode *)&fs_write_vnode,
	(op_remove_vnode *)&fs_remove_vnode,	//
	NULL, //(op_secure_vnode *)&fs_secure_vnode,	//
	(op_walk *)&fs_walk,
	(op_access *)&fs_access,
	(op_create *)&fs_create,
	(op_mkdir *)&fs_mkdir,
	(op_symlink *)&fs_symlink,
	(op_link *)&fs_link,					//
	(op_rename *)&fs_rename,
	(op_unlink *)&fs_unlink,
	(op_rmdir *)&fs_rmdir,
	(op_readlink *)&fs_readlink,
	(op_opendir *)&fs_opendir,
	(op_closedir *)&fs_closedir,
	(op_free_cookie *)&fs_free_dircookie,
	(op_rewinddir *)&fs_rewinddir,
	(op_readdir *)&fs_readdir,
	(op_open *)&fs_open,
	(op_close *)&fs_close,
	(op_free_cookie *)&fs_free_cookie,
	(op_read *)&fs_read,
	(op_write *)&fs_write,
	(op_readv *)&fs_readv,					//
	(op_writev *)&fs_writev,				//
	(op_ioctl *)&fs_ioctl,					//
	(op_setflags *)&fs_setflags,			//
	(op_rstat *)&fs_rstat,
	(op_wstat *)&fs_wstat,
	(op_fsync *)&fs_fsync,					//
	(op_initialize *)&fs_initialize,		//
	(op_mount *)&fs_mount,
	(op_unmount *)&fs_unmount,
	(op_sync *)&fs_sync,					//
	(op_rfsstat *)&fs_rfsstat,
	(op_wfsstat *)&fs_wfsstat,
	(op_select *)&fs_select,
	(op_deselect *)&fs_deselect,
	(op_open_indexdir *)&fs_open_indexdir,
	(op_close_indexdir *)&fs_close_indexdir,
	(op_free_cookie *)&fs_free_indexdircookie,
	(op_rewind_indexdir *)&fs_rewind_indexdir,
	(op_read_indexdir *)&fs_read_indexdir,
	(op_create_index *)&fs_create_index,
	(op_remove_index *)&fs_remove_index,
	(op_rename_index *)&fs_rename_index,
	(op_stat_index *)&fs_stat_index,
	NULL, //(op_open_attrdir *)&fs_open_attrdir,
	(op_close_attrdir *)&fs_close_attrdir,
	(op_free_cookie *)&fs_free_attrdircookie,
	(op_rewind_attrdir *)&fs_rewind_attrdir,
	(op_read_attrdir *)&fs_read_attrdir,
	NULL, //(op_write_attr *)&fs_writeattr,
	NULL, //(op_read_attr *)&fs_readattr,
	(op_remove_attr *)&fs_removeattr,
	(op_rename_attr *)&fs_rename_attr,
	NULL, //(op_stat_attr *)&fs_statattr,
	(op_open_query *)&fs_open_query,
	(op_close_query *)&fs_close_query,
	(op_free_cookie *)&fs_free_querycookie,
	(op_read_query *)&fs_read_query	
};
