#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <malloc.h>
#include <DataIO.h>
#include <lzo1x.h>

#include <Entry.h>

#include "global.h"
#include "client.h"
#include "nodemon.h"
#include "StringEx.h"
#include "StrTok.h"

// TClient

int lzo_compress(lzo_byte *in, lzo_uint in_len, lzo_byte *out, lzo_uint *out_len)
{
	lzo_byte *wrkmem = (lzo_bytep)malloc(LZO1X_1_MEM_COMPRESS);
	if (!wrkmem) return ENOMEM;
	if (lzo1x_1_compress(in, in_len, out, out_len, wrkmem) != LZO_E_OK) {
		free(wrkmem);
		return EIO;
	}
	free(wrkmem);
	return 0;
}

TClient::TClient(int clifd, struct sockaddr_in *cliaddr, const TConfig *config)
{
	fClientFd = clifd;
	fConfig = config;
	fFileHandle = NULL;
	
	if (clifd == -1) {
		fStdin = stdin;
		fStdout = stdout;
	} else {
		fClientAddr = *cliaddr;
		fStdin = fdopen(clifd, "r+");
		fStdout = fStdin;
	}
}

TClient::~TClient()
{
	if (fClientFd != -1) {
		fclose(fStdin);
		close(fClientFd);
	}
}

void TClient::run()
{
	if (login_auth()) {
		send_reply(NETBFS_OK, NETBFS_UNCOMPRESSED, 0, false, NULL);
		while(mainloop());
	} else {
		send_reply(NETBFS_EACCES, NETBFS_UNCOMPRESSED, 0, false, NULL);
	}
}

status_t TClient::send_reply(uint8 status, uint8 data_stat, uint32 data_length,
							uint8 continuation, const void *data)
{
	lzo_byte *data_buff = (lzo_byte *)&send_buff[sizeof(netbfs_reply_t)];
	netbfs_reply_t *reply = (netbfs_reply_t *)send_buff;
	lzo_uint compressed_size;
	uint32 send_buff_size;
	status_t err = B_OK;
	
	if (data_length > NETBFS_DATA_MAXSIZE) return EINVAL;
	
	if ((data_length == 0) || (data == NULL)) {
		data_length = 0;
		data_stat = NETBFS_UNCOMPRESSED;
	}
	
	switch(data_stat) {
		case NETBFS_UNCOMPRESSED:
			memcpy(data_buff, data, data_length);
			reply->r_data_stat = NETBFS_UNCOMPRESSED;
			reply->r_data_length = data_length;
			break;
		case NETBFS_LZO_COMPRESSED:
			compressed_size = sizeof(send_buff) - sizeof(netbfs_reply_t);
			if (lzo_compress((lzo_byte *)data, data_length, data_buff, &compressed_size) == 0) {
				if (compressed_size < data_length) {
					reply->r_data_stat = NETBFS_LZO_COMPRESSED;
					reply->r_data_length = compressed_size;
					break;
				}
			}
			// compress failed or incompressible
			memcpy(data_buff, data, data_length);
			reply->r_data_stat = NETBFS_UNCOMPRESSED;
			reply->r_data_length = data_length;
			break;
		default: return EINVAL;
	}
	
	reply->r_header = NETBFS_REPLY_HEADER;
	reply->r_status = status;
	reply->r_continuation = continuation;
	send_buff_size = sizeof(netbfs_reply_t) + reply->r_data_length;
	
	// send
	fwrite(send_buff, send_buff_size, 1, fStdout);
	
	return err;
}

status_t TClient::get_command(uint8 *command, uint32 *data_length, void **data)
{
	netbfs_cmd_t cmd;
	char *_cmd = (char *)&cmd;
	uint32 recvPos = 1, rsize;
	lzo_uint decomp_length = sizeof(decomp_temp);
	
	// read header
	cmd.c_header = 0;
	rsize = fread(&cmd, 1, 1, fStdin);
	if (feof(fStdin) || ferror(fStdin) || rsize == 0) return EIO;
	if (cmd.c_header != NETBFS_COMMAND_HEADER) return EINVAL;
	while(recvPos < sizeof(netbfs_cmd_t)) {
		rsize = fread(&_cmd[recvPos], 1, sizeof(netbfs_cmd_t) - recvPos, fStdin);
		recvPos += rsize;
		if (feof(fStdin) || ferror(fStdin) || rsize == 0) return EIO;
	}
	*command = cmd.c_command;
	
	// read data
	recvPos = 0;
	while(recvPos < cmd.c_data_length) {
		rsize = fread(&recv_buff[recvPos], 1, cmd.c_data_length - recvPos, fStdin);
		recvPos += rsize;
		if (feof(fStdin) || ferror(fStdin) || rsize == 0) return EIO;
	}
	
	switch(cmd.c_data_stat) {
		case NETBFS_UNCOMPRESSED:
			*data = recv_buff;
			*data_length = cmd.c_data_length;
			break;
		case NETBFS_LZO_COMPRESSED:
			if (lzo1x_decompress_safe((lzo_byte *)recv_buff, cmd.c_data_length, (lzo_byte *)decomp_temp, &decomp_length, 0) != 0) return EINVAL;
			*data = decomp_temp;
			*data_length = decomp_length;
			break;
		default: return EINVAL;
	}
	
	return B_OK;
}

int TClient::login_auth()
{
	LOG << "login" << endl;
	
	// login prompt 送信
	char *login_prompt = "netbfsd1.00";
	send_reply(NETBFS_OK, NETBFS_UNCOMPRESSED, strlen(login_prompt), 0, login_prompt);
	
	// 
	uint8 command;
	uint32 data_length;
	void *data;
	status_t err;
	if ((err = get_command(&command, &data_length, &data)) != B_OK) {
		LOG << "get_command error in login_auth(). (" << strerror(err) << ")" << endl;
		return false;
	}
	BString login_id((const char *)data, data_length);
	
	LOG << "Command received.";
	LOG << " command = " << (uint32)command;
	LOG << " data_length = " << data_length;
	LOG << " login_id = " << login_id.String() << endl;
	
	if (command != NETBFS_CMD_LOGIN) {
		LOG << "Invalid command. command = " << (uint32)command << endl;
		return false;
	}
	
	// 認証 / "username@password@sharename" 解析
	bool guest_ok = fConfig->IsGuestOk();
	TStrTok strtok(login_id.String(), "@", TOKEN_COUNT_NULLSTR);
	if (strtok.CountTokens() != 3) {
		LOG << "Invalid username. (" << login_id.String() << ")" << endl;
		return false;
	}
	const char *username = strtok.Token(0);
	const char *password = strtok.Token(1);
	const char *sharename = strtok.Token(2);
	
	//LOG << username << "@" << password << "@" << sharename << endl;
	
	const share_folder_t *share = fConfig->FindShare(sharename);
	if (share == NULL) {
		LOG << "Share name not found. (" << sharename << ")" << endl;
		return false;
	}
	
	// ゲストログイン許可？
	if (!share->guest_ok || !guest_ok) {
		// ゲスト不許可ならユーザー名とパスワードを確認
		LOG << "Checking username and password." << endl;
		const user_t *user = fConfig->FindUser(username);
		if (user == NULL) return false;
		if (strcmp(password, user->password->String()) != 0) return false;
		
		// 共有フォルダがユーザーを許可してるか確認
		TStrTok users(share->users->String(), ",", TOKEN_COUNT_NULLSTR);
		bool found = false;
		for(int32 i=0; i < users.CountTokens(); i++) {
			TStringEx str(users.Token(i));
			str.Trim();
			if (strcmp(str.String(), username) == 0) {
				found = true;
				break;
			}
		}
		if (!found) return false;
	} else {
		LOG << "Guest logged in." << endl;
	}
	
	fRootDir.SetTo(share->path->String());
	fWritable = fConfig->IsWritable() && share->writable;
	
	LOG << "Welcome!" << endl;
	
	return true;
}

int TClient::mainloop()
{
	uint8 command;
	uint32 data_length;
	char *data;
	bool cont = true;
	status_t err;
	
	if ((err = get_command(&command, &data_length, (void **)&data)) != B_OK) {
		// error
		LOG << "get_command() error in mainloop(). (" << strerror(err) << ")" << endl;
		return false;
	}
	
	switch(command) {
		case NETBFS_CMD_START_NODEMON: cmd_start_nodemon(data_length, data); break;
		case NETBFS_CMD_START_LIVEQUERY: cmd_start_livequery(data_length, data); break;
		case NETBFS_CMD_READDIR: cmd_readdir(data_length, data); break;
		case NETBFS_CMD_FOPEN: cmd_fopen(data_length, data); break;
		case NETBFS_CMD_FREAD: cmd_fread(data_length, data); break;
		default:
			LOG << "Unknown command. command = " << (uint32)command << "  data_length = " << data_length << endl;
	}	
	
	return cont;
}

// 相対パスから絶対パスを取得
int TClient::get_real_path(const char *path, BString *resolved_path)
{
	char resolved[MAXPATHLEN];
	if (realpath(path, resolved) == NULL) return -1;
	resolved_path->SetTo(resolved);
	return 0;
}


// netpath を realpath (絶対 path)に変換し root_dir 配下であるか調べる。
int TClient::netpath_to_realpath(const char *netpath, BString *real_path)
{
	int root_dir_len, status = -1;
	BString wk;
	
	// /path/dir/name -> /root_dir/path/dir/name 変換
	wk.SetTo(fRootDir);
	wk << "/" << netpath;
	
	// 相対パス表記をすべて解決
	if (get_real_path(wk.String(), real_path) != 0) return -1;
	
	// root_dir 配下であるか調べる
	root_dir_len = fRootDir.Length();
	if (real_path->Length() >= root_dir_len) {
		if (strncmp(real_path->String(), fRootDir.String(), root_dir_len) == 0) {
			status = 0;
			if (real_path->Length() > root_dir_len) {
				if (real_path->ByteAt(root_dir_len) != '/') {
					status = 1;
				}
			}
		}
	}
	return status;
}

// ファイル名を含む netfpath を realfpath に変換し、root_dir 配下であるか調べる。
int TClient::netfpath_to_realfpath(const char *netfpath, BString *realfpath)
{
	int status = 0;
	char *dirc, *basec, *bname, *dname;
	BString real_path;
	
	dirc = strdup(netfpath);
	basec = strdup(netfpath);
	dname = dirname(dirc);
	bname = basename(basec);
	if (netpath_to_realpath(dname, &real_path) != 0) {
		status = -1;
	} else {
		realfpath->SetTo(real_path.String());
		*realfpath << "/" << bname;
	}
	free(dirc);
	free(basec);
	return status;
}

// 指定の相対パス or 絶対パスが、root_dir 配下であるか調べる
int TClient::check_root_dir(const char *path)
{
	int result = 1;
	int root_dir_len;
	
	BString resolved_path;
	if (get_real_path(path, &resolved_path) == -1) {
		result = -1;
	} else {
		root_dir_len = fRootDir.Length();
		if (resolved_path.Length() >= root_dir_len) {
			if (strncmp(resolved_path.String(), fRootDir.String(), root_dir_len) == 0) {
				result = 0;
				if (resolved_path.Length() > root_dir_len) {
					if (resolved_path.ByteAt(root_dir_len) != '/') {
						result = 1;
					}
				}
			}
		}
	}
	return result;
}

// commands -----------------------------------------------------------------------------

void TClient::cmd_start_nodemon(uint32 data_length, char *data)
{
	
}

void TClient::cmd_start_livequery(uint32 data_length, char *data)
{
	
}

void TClient::cmd_readdir(uint32 data_length, char *data)
{
	fstat_t fstat;
	char _sendBuff[NETBFS_DATA_MAXSIZE];
	BMemoryIO sendBuff(_sendBuff, sizeof(_sendBuff));
	DIR *dp;
	struct stat st;
	struct dirent *de;
	BString netPath(data, data_length), realPath;
	
	if (netpath_to_realpath(netPath.String(), &realPath) != 0) {
		// No such file or directory
		send_reply(NETBFS_ENOENT, NETBFS_UNCOMPRESSED, 0, 0, NULL);
		return;
	}
	
	const char *baseDir = realPath.String();
	if ((dp = opendir(baseDir)) == NULL) {
		send_reply(NETBFS_ENOENT, NETBFS_UNCOMPRESSED, 0, 0, NULL);
		return;
	}
	
	while ((de = readdir(dp))) {
		// get fullpath name
		BString fullpath_name(baseDir);
		if (baseDir[strlen(baseDir) - 1] != '/') fullpath_name << "/";
		fullpath_name << de->d_name;
		
		// get file stat
		if (stat(fullpath_name.String(), &st) != 0) continue;
		
		// add entry
		BEntry entry(fullpath_name.String());
//		if (S_ISLNK(st.st_mode)) {
		if (entry.IsSymLink()) {
			// ignore link
			continue;
		}
		
		fstat.fs_ino = de->d_ino;
		fstat.fs_mode = st.st_mode;
		fstat.fs_size = st.st_size;
		fstat.fs_atime = st.st_atime;
		fstat.fs_mtime = st.st_mtime;
		fstat.fs_ctime = st.st_ctime;
		fstat.fs_crtime = st.st_crtime;
		fstat.fs_namelen = strlen(de->d_name);
		
		if ((sendBuff.Position() + sizeof(fstat) + fstat.fs_namelen) >= NETBFS_DATA_MAXSIZE) {
			// 送信バッファオーバー直前に送信
			send_reply(NETBFS_OK, NETBFS_LZO_COMPRESSED, sendBuff.Position() - 1, 1, _sendBuff);
			sendBuff.Seek(0, SEEK_SET);
		}
		
		sendBuff.Write(&fstat, sizeof(fstat));
		sendBuff.Write(de->d_name, fstat.fs_namelen);
	}
	
	send_reply(NETBFS_OK, NETBFS_LZO_COMPRESSED, sendBuff.Position(), 0, _sendBuff);
	
	closedir(dp);
}

// File Open
void TClient::cmd_fopen(uint32 data_length, char *data)
{
	if (fFileHandle) {
		fclose(fFileHandle);
		fFileHandle = NULL;
	}
	
	if (data_length < (sizeof(fopen_t) + 1)) {
		send_reply(NETBFS_EINVAL, NETBFS_UNCOMPRESSED, 0, 0, NULL);
		return;
	}
	
	fopen_t *fo = (fopen_t *)data;
	if (data_length != (sizeof(fopen_t) + fo->fo_namelen)) {
		send_reply(NETBFS_EINVAL, NETBFS_UNCOMPRESSED, 0, 0, NULL);
		return;
	}
	
	BString netPath(&data[sizeof(fopen_t)], fo->fo_namelen), realPath;
	
	if (netpath_to_realpath(netPath.String(), &realPath) != 0) {
		// No such file or directory
		send_reply(NETBFS_ENOENT, NETBFS_UNCOMPRESSED, 0, 0, NULL);
		return;
	}
	
	LOG << "fopen " << realPath.String() << endl;
	if ((fFileHandle = fopen(realPath.String(), "rb")) == NULL) {
		int err = errno;
		send_reply(err, NETBFS_UNCOMPRESSED, 0, 0, NULL);
		LOG << "fopen failed. " << strerror(err) << endl;
		return;
	}
	
	send_reply(NETBFS_OK, NETBFS_UNCOMPRESSED, 0, 0, NULL);
}

// File Read
void TClient::cmd_fread(uint32 data_length, char *data)
{
	if (fFileHandle == NULL) {
		send_reply(NETBFS_EINVAL, NETBFS_UNCOMPRESSED, 0, 0, NULL);
		return;
	}
	
	if (data_length != sizeof(fio_t)) {
		send_reply(NETBFS_EINVAL, NETBFS_UNCOMPRESSED, 0, 0, NULL);
		return;
	}
	
	fio_t *fio = (fio_t *)data;
	
	off_t off = fseek(fFileHandle, fio->fio_pos, SEEK_SET);
	if (off == -1) {
		send_reply(NETBFS_EINVAL, NETBFS_UNCOMPRESSED, 0, 0, NULL);
		return;
	}
	
	LOG << "fread " << fio->fio_pos << " " << fio->fio_len << " " << fFileHandle << endl;
	
	ssize_t len = 0;
	char buff[NETBFS_DATA_MAXSIZE];
	uint32 buffsize, readsize, totalsize = 0;
	uint8 continuation = 1;
	
	while (continuation > 0) {
		readsize = 0;
		buffsize = min_c(NETBFS_DATA_MAXSIZE, fio->fio_len - totalsize);
		while (readsize < buffsize) {
			LOG << "read " << buffsize << " " << readsize << endl;
			if ((len = fread(buff, 1, buffsize - readsize, fFileHandle)) == -1) {
				int err = errno;
				LOG << "error " << strerror(err) << endl;
				send_reply(err, NETBFS_UNCOMPRESSED, 0, 0, NULL);
				return;
			}
			LOG << "read len = " << len << endl;
			if (len == 0) break;
			readsize += len;
		}
		totalsize += readsize;
		if (len == 0) continuation = 0; else continuation = (totalsize < fio->fio_len) ? 1 : 0;
		send_reply(NETBFS_OK, /*NETBFS_UNCOMPRESSED*/NETBFS_LZO_COMPRESSED, readsize, continuation, buff);
		LOG << continuation << " " << totalsize << " " << fio->fio_len << endl;
	}
}
